#!/usr/bin/perl -w
# $Id$

use strict;
sub POE::Kernel::ASSERT_EVENT () { 1 }
# sub POE::Kernel::TRACE_FILES () { 1 }
# sub POE::Kernel::TRACE_REFCNT () { 1 }

use Test::More ( tests => 40 );
# use Religion::Package qw(1 1);

use POE::Component::IKC::Responder;
use POE::Component::IKC::Connection;
use POE qw(Kernel);
use POE::Wheel::Run;
use FindBin;
use Data::Denter;

use POE::Component::Daemon;


pass( 'loaded' );

sub DEBUG () { 0 }

my %OK;

POE::Component::IKC::Responder->spawn();

my $object = Test::Runner->spawn();

$poe_kernel->run();

pass( "Shutdown" );

delete $object->{bad_wheel};
delete $object->{server_wheel};

DEBUG and warn Denter $object;


############################################################################
package Test::Runner;

use strict;

use POE;
use Test::More;
use Config;
use Data::Denter;

my $perl;

BEGIN { 
    *DEBUG = \&main::DEBUG;
    $perl = $Config{perl5} || $Config{perlpath};

    if( $ENV{HARNESS_PERL_SWITCHES} ) {
        $perl .= " $ENV{HARNESS_PERL_SWITCHES}";
    }
 
}


###########################################################
sub spawn
{
    my($package)=@_;

    my $self = bless {}, $package;

    $self->{session} = POE::Session->create(
        inline_states=>{
            _start       => sub { $self->start( @_[ARG0..$#_] ) },
            do_next      => sub { $self->do_next( @_[ARG0..$#_] ) },
            start_server => sub { $self->start_server( @_[ARG0..$#_] ) },

            server_stdout => sub { $self->server_stdout( @_[ARG0..$#_] ) },
            server_stderr => sub { $self->server_stderr( @_[ARG0..$#_] ) },
            server_close  => sub { $self->server_close( @_[ARG0..$#_] ) },

            bad_stdout    => sub { $self->bad_stdout( @_[ARG0..$#_] ) },
            bad_stderr    => sub { $self->bad_stderr( @_[ARG0..$#_] ) },
            bad_close     => sub { $self->bad_close( @_[ARG0..$#_] ) },

            ERROR         => sub { $self->ERROR( @_[ARG0..$#_] ) },
            retryM        => sub { $self->retryM( @_[ARG0..$#_] ) },
            failM         => sub { $self->failM( @_[ARG0..$#_] ) },

            tcp_1         => sub { $self->tcp_1( @_[ARG0..$#_] ) },
            tcp_2         => sub { $self->tcp_2( @_[ARG0..$#_] ) },
            tcp_3         => sub { $self->tcp_3( @_[ARG0..$#_] ) },
            tcp_4         => sub { $self->tcp_4( @_[ARG0..$#_] ) },
            int_USR1 => sub {
                my( $kernel, $heap ) = @_[ KERNEL, HEAP ];
                my $ret = Daemon->__peek( 1 );
                $ret .= "heap = " . Denter $heap;
                warn "$$: $ret\n";
            }
        }
    )->ID;
    return $self;
}

###########################################################
sub start
{
    my( $self ) = @_;

    $self->{TODO} = [ 'unix-fail', 'tcp-fail', 
                      'launch-bad-server', 
                      'tcp-timeout', 'unix-timeout',
                      'launch-server', 'tcp', 'wait-for-it' ];

    $poe_kernel->alias_set( 'Runner' );

    $poe_kernel->yield( 'do_next' );

    $poe_kernel->call( IKC=>'monitor', '*', {
                    retry => 'retryM',
                    ERROR => 'ERROR',
                    fail  => 'failM',
                } );

    $poe_kernel->sig( USR1 => 'int_USR1' );

}

###########################################################
sub start_server
{
    my( $self ) = @_;
    DEBUG and warn "Launching test-server\n";

    $self->{server_wheel} = POE::Wheel::Run->new( 
        Program => "$perl -Iblib/lib $FindBin::Bin/test-server",
        ErrorEvent => 'error',
        StdoutEvent => 'server_stdout',
        StderrEvent => 'server_stderr',
        CloseEvent => 'server_close',
    );
}

sub server_stdout
{
    my( $self, $input, $wid ) = @_;

    DEBUG and warn "SERVER: $input\n";

    if( $input =~ /^PID=(\d+)/ ) {
        $self->{server_PID} = $1;
        $poe_kernel->yield( 'do_next' );
        END { kill 15, $self->{server_PID} if $self->{server_PID} };
    }
    elsif( $input =~ /\S/ ) {
        push @{ $self->{server_output} }, $input;
    }
}

sub server_stderr
{
    my( $self, $input, $wid ) = @_;

    DEBUG and warn "SERVER err: $input\n";
    push @{ $self->{server_error} }, $input;
}

sub server_close
{
    my( $self, $wid ) = @_;
    DEBUG and warn "SERVER_CLOSE";
    delete $self->{server_wheel};
    $self->{server_close}++;
}




###########################################################
sub start_bad
{
    my( $self ) = @_;
    DEBUG and warn "Launching test-bad-server\n";

    $self->{bad_wheel} = POE::Wheel::Run->new( 
        Program => "$perl -Iblib/lib $FindBin::Bin/test-bad-server",
        ErrorEvent => 'error',
        StdoutEvent => 'bad_stdout',
        StderrEvent => 'bad_stderr',
        CloseEvent => 'bad_close',
    );
}

sub bad_stdout
{
    my( $self, $input, $wid ) = @_;

    DEBUG and warn "BAD: $input\n";

    if( $input =~ /^PID=(\d+)/ ) {

        $poe_kernel->yield( 'do_next' );

        $self->{bad_PID} = $1;
        END { kill 15, $self->{server_PID} if $self->{server_PID} };
        END { kill 15, $self->{bad_PID} if $self->{bad_PID} };
        END { kill 15, $self->{bad_PID} if $self->{bad_PID} };
    }
    elsif( $input =~ /\S/ ) {
        push @{ $self->{bad_output} }, $input;
    }
}

sub bad_stderr
{
    my( $self, $input, $wid ) = @_;

    DEBUG and warn "BAD err: $input\n";
    push @{ $self->{bad_error} }, $input;
}

sub bad_close
{
    my( $self, $wid ) = @_;
    DEBUG and warn "BAD_CLOSE";
    delete $self->{bad_wheel};
    $self->{bad_close}++;
}





###########################################################
sub do_next
{
    my( $self ) = @_;

    my $type = shift @{$self->{TODO}};

    unless( $type ) {
        pass( "DONE" );
        diag( "kill -USR1 $$ # if this hangs" );

        # kill 15, delete $self->{server_PID} if $self->{server_PID};
        # kill 15, delete $self->{bad_PID} if $self->{bad_PID};

        # $poe_kernel->alias_remove( 'Runner' );
        $poe_kernel->post( IKC=>'shutdown' );
        return;
    }

    DEBUG and warn "\n\n\nDoing $type";

    if( $type eq 'unix-fail' ) {
        $self->unix_fail;
    }
    elsif( $type eq 'tcp-fail' ) {
        $self->tcp_fail;
    }
    elsif( $type eq 'launch-bad-server' ) {
        $self->start_bad;
    }
    elsif( $type eq 'tcp-timeout' ) {
        $self->tcp_timeout;
    }
    elsif( $type eq 'unix-timeout' ) {
        $self->unix_timeout;
    }
    elsif( $type eq 'launch-server' ) {
        kill 15, delete $self->{bad_PID} if $self->{bad_PID};

        $self->start_server;
    }
    elsif( $type eq 'tcp' ) {
        $self->tcp;
    }
    elsif( $type eq 'wait-for-it' ) {
        # ponder, ponder
    }
    else {
        die "What to do with type $type";
    }
}

sub retryM
{
    my( $self, $name, $rname ) = @_;
    DEBUG and warn "Retrying $name ($rname)";
    return unless $rname =~ /Client$/;
    $self->{retryM}++;
    ok( ( $self->{retryM} <= 5), "Allowed retries" );
}

sub failM
{
    my( $self, $name, $rname ) = @_;
    DEBUG and warn "Failed $name ($rname)";
    is( $self->{failM}, 1, "We wanted to mointor failure");
    $self->{failM}--;
}


#################################################################
sub unix_fail
{
    my( $self ) = @_;
    $self->{retry} = 0;
    $self->{retryM} = 0;
    $self->{fail} = 1;
    $self->{failM} = 1;

    my %params = (
            unix => ($ENV{TMPDIR}||$ENV{TEMP}||'/tmp').'/IKC-tests',
            aliases => [ 'Biff', 'Boff' ],
            connect_timeout => 2,
            reconnect_tries => 5,
            retry_delay     => 1,

            name            => 'UnixClient',

            on_connection   => sub { die "We don't want to be connected" },
            on_connect      => sub { die "We don't want to be connected" },
            on_error        => sub {
                my( $operation, $errstr, $errnum ) = @_;
                die "$operation error: $errstr ($errnum)";
            },
            on_retry        => sub {
                $self->{retry}++;
                ok( ( $self->{retry} <= 5 ), "5 retries" )
                    or die "To many retries";
            },
            on_fail         => sub {
                is( $self->{fail}, 1, "We wanted to fail");
                $self->{fail}--;
                DEBUG and warn "NEXT";
                $poe_kernel->post( $self->{session}, 'do_next' );
            },
        );
    DEBUG and warn "Unix client $params{unix} will fail\n";
    POE::Component::IKC::Connection->spawn( %params );
}

#################################################################
sub tcp_fail
{
    my( $self ) = @_;

    $self->{retry} = 0;
    $self->{retryM} = 0;
    $self->{fail} = 1;
    $self->{failM} = 1;

    my %params = (
            port => 1338,

            connect_timeout => 10,
            reconnect_tries => 2,

            name            => 'TcpClient',

            on_connection   => sub { die "We don't want to be connected" },
            on_connect      => sub { die "We don't want to be connected" },
            on_error        => sub {
                my( $operation, $errstr, $errnum ) = @_;
                die "$operation error: $errstr ($errnum)";
            },
            on_retry        => sub {
                $self->{retry}++;
                ok( ( $self->{retry} <= 2 ), "2 retries" )
                    or die "To many retries";
            },
            on_fail         => sub {
                is( $self->{fail}, 1, "We wanted to fail");
                $self->{fail}--;

                $poe_kernel->post( $self->{session}, 'do_next' );
            },
        );
    DEBUG and warn "TCP client :$params{port} will fail\n";

    POE::Component::IKC::Connection->spawn( %params );
}

#################################################################
sub tcp_timeout
{
    my( $self ) = @_;

    $self->{retry} = 0;
    $self->{retryM} = 0;
    $self->{fail} = 1;
    $self->{failM} = 1;

    my %params = (
            port => 9191,

            connect_timeout => 5,
            reconnect_tries => 2,
            register_timeout => 2,
            retry_delay     => 0,

            name            => 'TcpClient',

            on_connection   => sub { 
                # Gets called twice
                ok(1, "Connected TCP" );
            },
            on_connect      => sub { die "We don't want to be connected" },
            on_error        => sub {
                my( $operation, $errstr, $errnum ) = @_;
                die "$operation error: $errstr ($errnum)";
            },
            on_retry        => sub {
                $self->{retry}++;
                ok( ( $self->{retry} <= 2 ), "2 retries" )
                    or die "To many retries";
            },
            on_fail         => sub {
                is( $self->{fail}, 1, "We wanted to fail");
                $self->{fail}--;

                $poe_kernel->post( $self->{session}, 'do_next' );
            },
        );
    DEBUG and warn "TCP client :$params{port} will fail\n";

    POE::Component::IKC::Connection->spawn( %params );
}

#################################################################
sub unix_timeout
{
    my( $self ) = @_;

    $self->{retry} = 0;
    $self->{retryM} = 0;
    $self->{fail} = 1;
    $self->{failM} = 1;

    my %params = (
            unix => ($ENV{TMPDIR}||$ENV{TEMP}||'/tmp').'/IKC-tests',

            reconnect_tries => 2,
            register_timeout => 2,
            retry_delay     => 0,

            name            => 'UnixClient',

            on_connection   => sub { 
                # Gets called twice
                ok(1, "Connected UNIX" );
            },
            on_connect      => sub { die "We don't want to be connected" },
            on_error        => sub {
                my( $operation, $errstr, $errnum ) = @_;
                die "$operation error: $errstr ($errnum)";
            },
            on_retry        => sub {
                $self->{retry}++;
                ok( ( $self->{retry} <= 2 ), "2 retries" )
                    or die "To many retries";
            },
            on_fail         => sub {
                is( $self->{fail}, 1, "We wanted to fail");
                $self->{fail}--;

                $poe_kernel->post( $self->{session}, 'do_next' );
            },
        );
    DEBUG and warn "Unix client $params{unix} will fail\n";

    POE::Component::IKC::Connection->spawn( %params );
}

#################################################################
sub tcp
{
    my( $self ) = @_;

    $self->{retry} = 0;
    $self->{retryM} = 0;
    $self->{fail} = 0;
    $self->{failM} = 0;

    my %params = (
            port => 1338,

            connect_timeout => 5,
            reconnect_tries => 0,
            register_timeout => 5,
            retry_delay     => 10,

            name            => 'TcpClient',

            on_connection   => sub { ok(1, "Connected TCP" ); },
            on_connect      => sub { 
                ok(1, "Connect" );
                unless( $self->{tcp_1}++ ) {
                    $poe_kernel->post( $self->{session} => 'tcp_1', @_ );
                }
            },
            on_disconnect   => sub {
                pass( "Server disconnected" );
                if( $self->{tcp_3} == 0 ) {
                    $poe_kernel->post( $self->{session} => 'tcp_3', @_ );
                }
                elsif( $self->{tcp_3} == 1) {
                    $poe_kernel->post( $self->{session} => 'do_next' );
                }

                $self->{tcp_3}++;
            },
            on_error        => sub {
                my( $operation, $errstr, $errnum ) = @_;
                die "$operation error: $errstr ($errnum)";
            },
            on_retry        => sub {
                $self->{retry}++;
                ok( ( $self->{retry} <= 2 ), "2 retries" )
                    or die "To many retries";
            },
            on_fail         => sub {
                fail( "We aren't allowed to fail" );
            },
        );
    DEBUG and warn "TCP client :$params{port} will succeed\n";

    POE::Component::IKC::Connection->spawn( %params );
}

###########################################
sub tcp_1
{
    my( $self, $rid ) = @_;
    ok( $rid, "Got a remote kernel ID via on_connect" );
    DEBUG and warn "Connected to $rid";

    $self->{remote_kernel} = $rid;
    $poe_kernel->call( IKC=>'publish', Runner => [ qw( tcp_2 ) ] );

    $poe_kernel->post( IKC=>'call', 
                        "poe://$rid/test/called", "called", 
                        "poe://TcpClient/Runner/tcp_2" );
}

###########################################
sub tcp_2
{
    my( $self, $answer ) = @_;
    DEBUG and warn "Answer is $answer";

    is( $answer, 17, "Answer is 17!" );

    # now we close the other side down, which will cause a retry
    $poe_kernel->post( IKC=>'post', 
                        "poe://$self->{remote_kernel}/test/done" );
}

###########################################
sub tcp_3
{
    my( $self, $rid ) = @_;
    DEBUG and warn "Disconnected from $rid";

    ok( $rid, "Got a remote kernel ID via on_disconnect" );
    is( $self->{remote_kernel}, $rid, 
                "Same one as do_connect told us about before" );

    $poe_kernel->call( IKC=>'publish', Runner => [ qw( tcp_4 ) ] );

    $self->{remote_kernel} = $rid;

    # Send a request.  This will be held by IKC::Connection until the link
    # opens up
    $poe_kernel->post( IKC=>'call', 
                        "poe://Inet/test/called", "called", 
                        "poe://TcpClient/Runner/tcp_4" );
    $poe_kernel->yield( 'start_server' );
}

###########################################
sub tcp_4
{
    my( $self, $answer ) = @_;
    DEBUG and warn "Answer is $answer";

    is( $answer, 17, "Answer is 17!  HOT DAMN!" );

    $poe_kernel->post( IKC=>'post', 
                        "poe://$self->{remote_kernel}/test/done" );
#    $poe_kernel->yield( 'do_next' );
}



__END__

$Log$
