#!/usr/bin/perl -w
# $Id$

use strict;

# sub POE::Kernel::ASSERT_EVENTS { 1 }
# sub POE::Kernel::TRACE_REFCNT { 1 }

use Test::More ( tests => 18 );

use POE::Component::IKC::ClientLite;
use POE::Component::IKC::Server;
use POE::Component::IKC::Responder;
use Data::Dumper;
use POE qw(Kernel);

pass( "loaded" );

sub DEBUG () {0}

# try finding a freezer
my $p = POE::Component::IKC::ClientLite::_default_freezer();

ok( $p, "Found a freezer" );

# try loading freezer
my($f, $t)=
    POE::Component::IKC::ClientLite::_get_freezer('POE::Component::IKC::Freezer');
ok( ($f and $t), "Loaded a freezer" );

POE::Component::IKC::Responder->spawn;

POE::Component::IKC::Server->spawn(
        port=>1337,
        name=>'Inet',
        aliases=>[qw(Ikc)],
    );

DEBUG and print "Test server $$\n";
Test::Server->spawn();

$poe_kernel->run();

pass( "survived" );


############################################################################
package Test::Server;
use strict;
use Config;
use POE::Session;

use Test::More;

BEGIN {
    *DEBUG=\&::DEBUG;
}

###########################################################
sub spawn
{
    my($package)=@_;
    POE::Session->create(
#         args=>[$qref],
        package_states=>[
            $package=>[qw(_start _stop fetchQ add_1 add_n here
                        lite_register lite_unregister
                        shutdown do_child timeout do_next
                        )],
        ],
    );
}

###########################################################
sub _start
{
#    use Data::Denter;
#    die Denter "KERNEL is ". 0+KERNEL, \@_;
    my($kernel, $heap)=@_[KERNEL, HEAP, ARG0];
    DEBUG and warn "Test server: _start\n";
    pass( "_start" );

    $kernel->alias_set('test');
    $kernel->call(IKC=>'publish',  test=>[qw(fetchQ add_1 here)]);

    $kernel->post(IKC=>'monitor', 
       'LiteClient'=>{
            register=>'lite_register',
            unregister=>'lite_unregister'
        }); 
    $kernel->post(IKC=>'monitor', 
       'LiteClient0'=>{
            register=>'lite_register',
            unregister=>'lite_unregister'
        });
    $kernel->post(IKC=>'monitor', '*'=>{shutdown=>'shutdown'});

    $heap->{todo} = [ 'wait', 'lite', 'wait', 'lite0' ];

    $kernel->yield( 'do_next' );
}

my $count;

sub do_next
{
    my( $kernel, $heap ) = @_[ KERNEL, HEAP ];

    my $step = shift @{ $heap->{todo} };

    unless( $step ) {
        DEBUG and warn "DONE";
        $kernel->post(IKC=>'shutdown');
        return;
    }

    DEBUG and warn "\n\nDoing $step";

    
    if( $step eq 'wait' ) {
        $kernel->delay( do_next=>1 );
    }
    elsif( $step eq 'lite' ) {
        $count=0;
        $kernel->yield( do_child => 'lite' );
    }
    elsif( $step eq 'lite0' ) {
        diag( "You will see 2 warnings about falling back to version 0" );
        $kernel->yield( do_child => 'lite', 0 );
    }
    else {
        die "What does step $step do?";
    }
}

###########################################################
sub do_child
{
    my($kernel, $type, $protocol)=@_[KERNEL, ARG0, ARG1];
    my $pid=fork();
    die "Can't fork: $!\n" unless defined $pid;
    if($pid) {          # parent
        $kernel->delay(timeout=>60);
        return;
    }
    my $exec="$Config{perlpath} -I./blib/arch -I./blib/lib -I$Config{archlib} -I$Config{privlib} t/test-$type $protocol";

    if( $ENV{HARNESS_PERL_SWITCHES} ) {
        $exec =~ s/-I/$ENV{HARNESS_PERL_SWITCHES} -I/;
    }

    DEBUG and warn "Running $exec";
    exec $exec;
    die "Couldn't exec $exec: $!\n";
}


###########################################################
sub _stop
{
    my($kernel, $heap)=@_[KERNEL, HEAP, ARG0];
    DEBUG and warn "Test server: _stop\n";
    pass( "_stop" );
}


###########################################################
sub lite_register
{
    my($kernel, $heap, $name, $alias, $is_alias,
                            )=@_[KERNEL, HEAP, ARG0, ARG1, ARG2];
    $count++;
    DEBUG and warn "Test server: lite_register count=$count\n";

    if( $count == 1 ) {
        is( $name, 'LiteClient', "Connected from LiteClient" );
    }
    elsif( $count == 5 ) {
        is( $name, 'LiteClient0', "Connected from LiteClient0" );
    }
}

###########################################################
sub lite_unregister
{
    my($kernel, $heap, $name, $alias, $is_alias,
                            )=@_[KERNEL, HEAP, ARG0, ARG1, ARG2];
    $count++;
    DEBUG and warn "Test server: lite_unregister count=$count\n";

    my $done = 0;
    if( $count==2 or $count==4 ) {
        is($name, 'LiteClient', "Disconnected from LiteClient" );
        $done = ($count == 4);
    }
    elsif( $count==6 or $count==8 ) {
        is($name, 'LiteClient0', "Disconnected from LiteClient0" );
        $done = ($count==8);
    }

    return unless $done;

    $kernel->delay('timeout');          # set in do_child
    $kernel->yield( 'do_next' );
}

###########################################################
sub shutdown
{
    my($kernel)=$_[KERNEL];
    $kernel->alias_remove('test');
    DEBUG and warn "Test server: shutdown\n";
#    use YAML qw(Dump);
#    use Data::Dumper;
#    warn Dumper $kernel;
}
###########################################################
sub fetchQ
{
    my($kernel, $heap)=@_[KERNEL, HEAP];
    pass( "fetchQ" );
    return 7
}

###########################################################
sub add_1
{
    my($kernel, $heap, $args)=@_[KERNEL, HEAP, ARG0];
    DEBUG and warn "$$: add_1";
    my($n, $pb)=@$args;
    DEBUG and warn "$$: foo $n";
    is( $n, 7, "Fetched 7" );
    $kernel->yield('add_n', $n, 1, $pb);
}

###########################################################
sub add_n
{
    my($kernel, $n, $q, $pb)=@_[KERNEL, ARG0, ARG1, ARG2];
    DEBUG and warn "$$: add_n $n+$q";
    $kernel->post(IKC=>'post', $pb=>$n+$q);
}

###########################################################
sub here
{
    my($kernel, $n)=@_[KERNEL, ARG0];
    DEBUG and warn "$$: here n=$n";
    pass( "Got here" );
}

###########################################################
sub timeout
{
    my($kernel)=$_[KERNEL];
    warn "Test server: Timedout waiting for child process.\n";
    $kernel->post(IKC=>'shutdown');
}




