#!/usr/bin/perl -w
# $Id4

use strict;

use Test::More ( tests => 6 );

use POE::Component::IKC::Server;
use POE::Component::IKC::Channel;
use POE::Component::IKC::Client;
use POE qw(Kernel);

pass( "loaded" );

sub DEBUG () {0}

POE::Component::IKC::Server->spawn(
        port=>1338,
        name=>'Inet',
        aliases=>[qw(Ikc)],
    );

Test::Server->spawn();
Test::Client->spawn();

$poe_kernel->run();

pass( "survived" );



############################################################################
package Test::Server;
use strict;
use POE::Session;

use Test::More;

BEGIN {
    *DEBUG=\&::DEBUG;
}

###########################################################
sub spawn
{
    my($package)=@_;
    POE::Session->create(
#         args=>[$qref],
        package_states=>[
            $package=>[qw(_start _stop called shutdown)],
        ],
    );
}

###########################################################
sub _start
{
    my($kernel, $heap)=@_[KERNEL, HEAP, ARG0];
    DEBUG and warn "Server: _start\n";
    pass( "_start" );

    $kernel->alias_set('test');
    $kernel->call(IKC=>'publish',  test=>[qw(called)]);

    $kernel->post(IKC=>'monitor', '*'=>{shutdown=>'shutdown'});
}

###########################################################
sub _stop
{
    my($kernel, $heap)=@_[KERNEL, HEAP, ARG0];
    DEBUG and warn "Server: _stop\n";
}


###########################################################
sub called
{
    my($kernel, $heap)=@_[KERNEL, HEAP, ARG0];
    DEBUG and warn "Server: _stop\n";
    pass( "called" );
    return 4;
}


###########################################################
sub shutdown
{
    my($kernel, $heap, $name, $alias, $is_alias, 
                            )=@_[KERNEL, HEAP, ARG0, ARG1, ARG2];
    DEBUG and warn "Server: shutdown\n";
    $kernel->alias_remove('test');
    pass( "shutdown" );
}

############################################################################
package Test::Client;
use strict;
use POE::Session;

use Test::More;

BEGIN {
    *DEBUG=\&::DEBUG;
}


###########################################################
sub spawn
{
    my($package)=@_;
    POE::Session->create(
        package_states=>[
            $package=>[qw(_start callback)],
        ],
    );
}

###########################################################
sub _start
{
    my($kernel, $heap)=@_[KERNEL, HEAP, ARG0];
    DEBUG and warn "Client: _start\n";

    $kernel->alias_set('TC');

    $kernel->post(IKC=>'call', "poe://Inet/test/called", '', "poe:callback");
}

###########################################################
sub callback
{
    my($kernel, $heap, $n)=@_[KERNEL, HEAP, ARG0];
    DEBUG and warn "Client: callback\n";
    is( $n, 4, "Callback with 4" );
    $kernel->alias_remove('TC');
    $kernel->post(IKC=>'shutdown');
}


